#!/usr/local/bin/perl
######################################################################
#
# speedstats.pl
# 
# a qconsole log parser for TF
# by Erik Selberg <speed@cs.washington.edu>
# Started 10/13/97
#
# Map bugs and other weirdness
#
# 32smooth
#   reports "xxx got the enemy key" which is fine, except that 
#   it doesn't say if the person recording the log got the key.
#   It also doesn't print a newline after "Blah got the flag!" thus
#   that trimming code down there. 
#
# TODO
#
# switch between .html and .htm for M$ weenies
# better control of filenames
#


require 5.0001;
use strict;
use IO::File;

my $version = "1.4";
my $testing = 0;

# for sorting routines

######################################################################
#
# Things that look a lot like constants
#

my $ARGS = join(" ", @ARGV);
my $obitfile = "obits.txt";
my $logfile = "qconsole.log";
my $outfile = "";
my $maketable = "http://huskysearch.cs.washington.edu/fifth/maketable.js";
my $speedstyle = "http://huskysearch.cs.washington.edu/fifth/speedstyle.css";
my $statsdir = ".";

my $cmdline;

my $classfile = "weapclass.txt";
my $mode = "fortress";
my $modeline;

my $serverlog = 0;
my $noprivate = 1;
my $spectalk = 1;

# if there aren't enough kills, don't bother printing anything
my $kill_threshhold = 20;
my $teamkill_threshhold = 10;

# older font colors
# my $blue = "#8080ff";
# my $orange = "#ff8000";
# my $gold = "#ff8000";

# >V< Someone's font colors
my $blue = "#8080ff";
my $orange = "#d2691e";
my $gold = "#b8860b";

# use spaces in the tails to distinguish them. cheap hack, I know
my $bfonth = "<font color=$blue>";
my $bfontt = "</font >";

my $ofonth = "<font color=$orange>";
my $ofontt = "</font  >";

my $gfonth = "<font color=$gold>";
my $gfontt = "</font   >";

my $layerh = "<layer TOP=&{window.document.layers[window.document.layers.length - 1].top + window.document.layers[window.document.layers.length - 1].document.height};>\n";
my $layert = "</layer>\n";

######################################################################
# 
# experimental O(n) parsing stuff

my $massivere;
my %massivehash;
my $mmatch;
my $ignorere;

my @massiverelist;
my @massivereidx;

my $talkre;

######################################################################
#
# stuff that doesn't get cleared when we create a new file
#

use Cwd;

my $cwd = cwd();

sub find_real_dir {
  my ($d) = @_;
  
  return "." if ($d !~ /\//);
  
  my $cwd = cwd();
  my $f;
  
  while (-l $d) {
    ($f = $d) =~ s|/[^\/]+$||;
    chdir $f;
    $d = readlink($d);
  }
  chdir($cwd);
  return $d;
}

my $pdir;
my $pname;
my $real_prog = find_real_dir($0);
($pdir = $real_prog) =~ s|/[^\/]+$||;
($pname = $real_prog) =~ s|^.*/([^\/]+)$|$1|;

print "Local dir is $cwd\n" if ($testing);
print "Real_prog is: $real_prog\n" if ($testing);

my $quiet;

my $title = "";
my $spacedtitle = "";
my $date = get_file_cdate($logfile);
my $thissess = 0;

my @typeidx;
my @regexps;
my $numregexps;
my @goodplayers;

my $templatedir = "$pdir/templates";

my $indiv_template;
my $table_template;
my $head_template;
my $foot_template;
my $playerlist_template;
my $playerkills_template;
my $playersum_template;
my $weaponkills_template;
my $weapondeaths_template;
my $classkills_template;
my $result_template;
my $frame_template;
my $nav_template;
my $mm1_template;

my %defteams = ( "blue" => 1,
		 "red" => 2,
		 "yell" => 3,
		 "gren" => 4
	       );

my @teamcolors = ("blue", "red", "yellow", "green");

my $MAINOUT;

my %classes;
my %classnames;
my %weapons;

my $logline = 0;

sub get_timestr {
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
      = localtime(time);
    return sprintf("%d-%d_%d%s", $mon + 1, $mday + 1, $year,
		   ($testing) ? "-$hour:$min:$sec" : "");
}
my $nowstr = get_timestr();


######################################################################
#
# stuff that does
#

my %players;
my %killedby;
my %killedbyweap;
my %killedwithweap;
my %killedbyweapclass;
my %killedwithweapclass;
my %kills;
my %deaths;
my %trap;
my %suicides;
my %monster;
my %klutz;
my %chatty;
my %overflowed;
my %dropped;
my %award;
my %weaponkills;
my %killedbuddy;

my %bioinfect;
my %biocured;
my %biodiseased;

my %enter;
my %notoss;
my $detpacks;
my @sections;
my %build;
my %lastclass;
my $parsed;
my %unknown_weap;
my %unknown_targ;
my %flagtouch;
my %flagcaps;
my %flagdrops;
my %flagreturns;
my $resultsin;
my $maxteam = 2;   # is this a bad default?
my $team;
my $winner;
my $loser;
my %results_players;
my %results_score;
my %results_frags;
my %results_fragsdropped;
my %possible_clans;
my @mm1log;
my %munged_names;
my %original_names;
my %clean_names;
my %bonus;
my %assist;
my %defend;
my %player_clans;
my %team;
my %clanlist;
my %team2clan;
my %clan2team;
my $clans_detected;
my %setclan;
my %caught_spy;

my $bonus_next_kill = 0;
my $last_killer;

# unique ID for name; this is used to determine aliases
my %name_id;
my %id_name;

sub clear_all {
    %name_id = ();
    %id_name = ();
    %caught_spy = ();
    %setclan = ();
    $clans_detected = 0;
    %team2clan = ();
    %clan2team = ();

    %clanlist = ();
    %players = ();
    %killedby = ();
  %killedbyweap = ();
  %killedwithweap = ();
  %killedbyweapclass = ();
  %killedwithweapclass = ();
  %kills = ();
  %deaths = ();
  %trap = ();
  %suicides = ();
  %monster = ();
  %klutz = ();
  %chatty = ();
  %overflowed = ();
  %dropped = ();
  %award = ();
  %weaponkills = ();
  %killedbuddy = ();

    %bioinfect = ();
    %biocured = ();
    %biodiseased = ();

  %enter = ();
  %notoss = ();
  $detpacks = 0;
  @sections = ();
  %build = ();
  %lastclass = ();
  $parsed = 0;
  %unknown_weap = ();
  %unknown_targ = ();
  %flagdrops = ();
  %flagtouch = ();
  %flagcaps = ();
  %flagreturns = ();
  $resultsin = 0;
    $maxteam = 2;
  $team = 0;
  $winner = 0;
  $loser = 0;
  %results_players = ();
  %results_score = ();
  %results_frags = ();
  %results_fragsdropped = ();
  %possible_clans = ();
  @mm1log = ();
  %munged_names = ();
  %original_names = ();
  %clean_names = ();
  $title = "";
  %bonus = ();
  $bonus_next_kill = 0;
  %assist = ();
  %defend = ();
  %player_clans = ();
  %team = ();
}


sub read_template_file {
  my ($tempfile) = @_;
  my $template_str = "";
  
  # yeah, dumb way... but winblows doesn't like `cat ...`
  open(INFILE, "<$templatedir/$tempfile");
  while (<INFILE>) {
    $template_str .= $_;
  }
  close(INFILE);
  
  my $template = $template_str;
  # make sure that sprintf doesn't barf on quotes or %'s
  $template =~ s/"/\\"/g;
  $template =~ s/%/%%/g;
  return $template;
}

# perhaps make the template file names part of some neato cmd line option?
sub read_templates {
  $indiv_template = read_template_file("indiv_template.html");
  $table_template = read_template_file("table_template.html");
  $head_template = read_template_file("head_template.html");
  $foot_template = read_template_file("foot_template.html");
  $playerlist_template = read_template_file("playerlist_template.html");
  $playerkills_template = read_template_file("playerkills_template.html");
  $playersum_template = read_template_file("playersum_template.html");
  $weaponkills_template = read_template_file("weaponkills_template.html");
  $weapondeaths_template = read_template_file("weapondeaths_template.html");
  $classkills_template = read_template_file("classkills_template.html");
  $result_template = read_template_file("result_template.html");
  $frame_template = read_template_file("frame_template.html");
  $nav_template = read_template_file("nav_template.html");
  $mm1_template = read_template_file("mm1_template.html");
}

sub usage {
  print <<EOM
  Usage: $0 [-obits <obitfile>] [-log [<logfile> | -]] [-mode <mode>] [-out outfile] [-quiet | -q] [-maketable <maketable.js url>] [-serverlog] [-dir statsdir] [-nofilter] [-nospec]
EOM
    ;
  exit;
}

sub main {
    print "Running as: $0 ", join(" ", @ARGV), "\n";

    $cmdline = "$0 " . join(" ", @ARGV);

  while (@ARGV) {
    $_ = shift @ARGV;
    (/quiet/ || /-q/) && do {
      $quiet++;
      next;
    };
    /nospec/ && do {
	$spectalk = 0;
	next;
    };
    /nofilter/ && do {
	$noprivate = 0;
	next;
    };
    (/testing/) && do {
      $testing = shift @ARGV;
      next;
    };
    (/serverlog/) && do {
	$serverlog = 1;
	next;
    };
    (/kills/) && do {
	$kill_threshhold = shift @ARGV;
	next;
    };
    (/obits/) && do {
      $obitfile = shift @ARGV;
      next;
    };
    (/log/ || /-l/) && do {
      $logfile = shift @ARGV;
      $date = get_file_cdate($logfile);
      next;
    };
    /mode/ && do {
	if (-e "$templatedir/${mode}_$classfile") {
	    $mode = shift @ARGV;
	    read_classes("${mode}_$classfile");
	} else {
	    print "Error, couldn't switch to mode $mode (no ${mode}_$classfile in $templatedir!";
	    shift @ARGV;
	}
	next;
    };
    (/out/ || /-o/) && do {
      $outfile = shift @ARGV;
      next;
    };
    /maketable/ && do {
	$maketable = shift @ARGV;
	next;
    };
    /style/ && do {
	$speedstyle = shift @ARGV;
	next;
    };
    /dir/ && do {
       my $newdir;
       $newdir = shift @ARGV;
       if (! -d $newdir) {
           print "Woops, $newdir isn't a directory... still using $statsdir\n";
       } else {
	   $statsdir = $newdir;
       }
       next;
    };
    print "Unknown option: $_\n";
    usage();
  }
  
  # if we have an outfile, open and select it
  # otherwise, use STDOUT
  if ("$outfile" ne "") {
    open('MAINOUT', ">$outfile") || die("Couldn't open outfile $outfile!");
    $MAINOUT = 'MAINOUT';
  } else {
    $MAINOUT = 'STDOUT';
  }
  select $MAINOUT;
  
  read_obits($obitfile);
  read_templates();
  read_classes("${mode}_${classfile}");
  parse_logfile($logfile);
  
  if ($testing =~ /[ac]/) {
    print "\nPlayers:\n";
    foreach (keys(%players)) {
      print $clean_names{$_}, "\n";
    }

  }
}

sub get_file_cdate {
  my ($fname) = @_;
  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst);

  if ($fname && -f $fname) {
      my @stats = stat($fname);
      my $fcdate = $stats[9];
      ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
	  = localtime($fcdate);
  } else {
      ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
	  = localtime(time);
  }

  return sprintf("%d/%d/%d", $mon + 1, $mday, $year);
}

sub get_matches {
  my($rfunc, @vals) = @_;
  my @matches;
  
  return () unless (@vals);
  
  my $key = &$rfunc($vals[0]);
  my $i = 0;
  while (&$rfunc($vals[$i]) == $key) {
    push(@matches, $vals[$i++]);
  }
  return @matches;
}

sub member {
  my($str, @l) = @_;
  
  foreach (@l) {
    return 1 if ($str eq $_);
  }
  return 0;
}

######################################################################
#
# For TF only, guesses class based on various heuristics.
# If we KNOW the class via the weapon, we're done.
# If we DON'T know the class, put 'the stats in limbo until we come across
# an entry we DO know.
#
# new for TF 2.8: we have class-specific messages (woot!) so lots of this is
# now obsolete.

sub guess_class {
  my($weap, $player, $targ) = @_;
  
  my @possibilities = split (/\s/, $classes{$weap});
  
  if (@possibilities == ()) {
    print "ERROR! guess_class($weap, ", $clean_names{$player}, ", ", $clean_names{$targ}, ") has no possibilities from $classes{$weap}\n";
    return "ERROR";
  } elsif (@possibilities == 1) {
    $lastclass{$player} = $possibilities[0];
    unify_unknown_kills($possibilities[0], $player);
    return $possibilities[0];
  }
  
  # ok... if the weapon is consistant with the LAST class the player
  # played, we'll assume there hasn't been a change.
  
  return $lastclass{$player} if (member($lastclass{$player}, 
					@possibilities));
  
  # ok... we don't know what class the player is. :( So we store this entry,
  # and hope we can narrow it down.
  
  $unknown_weap{$player} .= "$weap ";
  $unknown_targ{$player} .= "$targ ";
  print "REDRUM! Couldn't figure out class for ", $clean_names{$player}, " and $weap (lastclass was $lastclass{$player}, deferring\n" if ($testing =~ /[ah]/);
  return "UNKNOWN";
  
}

sub unify_unknown_kills {
  my($weap, $player) = @_;
  my $class;
  my @possibilities;
  
  my @unknowns = split(" ", $unknown_weap{$player});
  my @unknown_targs = split(" ", $unknown_targ{$player});
  my $i = 0;
  foreach (@unknowns) {
    @possibilities = split " ", $classes{$_};
    if (member($lastclass{$player}, @possibilities)) {
      $class = $lastclass{$player};
    } else {			# crap... we got an outlyer. :(
      # a good solution would be to save this until we've seen
      # all the player's classes. 
      # but for now, just pick the first. :)
      $class = $possibilities[0];
      print "REDRUM! Couldn't figure out class for ", $clean_names{$player}, " and $_ (lastclass was $lastclass{$player}, picking $class\n" if ($testing =~ /[ah]/);
  }
    #	print "REDRUM! Using $class for ", $clean_names{$player},"\n";
    $killedbyweapclass{$unknown_targs[$i], $_, $class}++;
    $killedwithweapclass{$player, $_, $class}++;
    $i++;
}
  $unknown_weap{$player} = "";
  $unknown_targ{$player} = "";
}

sub generate_title {
    my ($t) = @_;

    return "<table width=\"100%%\"><tr><td bgcolor=\"$blue\" align=center>\n" .
	"<font size=5><font color=black>" .
	    "<b>$t</b></font ></font>\n" .
		"</td></tr></table>\n";
}

sub print_header {
  print eval "sprintf \"$head_template\n\"";
}

sub print_frame {
  print eval "sprintf \"$frame_template\n\"";
}

sub print_nav {
  my $numplayers = @goodplayers;
 
  my $playerlist = join(", ", map {sprintf("\"%s\"", clean_name($_, 1));} @goodplayers);
  my $playeropts = join(", ", map {sprintf("\"%s\"", $_);} @goodplayers);

  my $player_accomps_opts = join(", ", map {s/\s/_/g; sprintf("\"%s\"", $_);} sort @sections);
  my $player_accomps = join(", ", map {s/\s/_/g; sprintf("\"%s\"", $_);} sort @sections);


  my $player_classes = join(", ", map {sprintf("\"%s\"", $_);} sort keys(%classnames));

  print eval "sprintf \"$nav_template\n\"";
}


sub print_matchresults {
  my $i;
  my $restables;
  my $stitle = generate_title("Match Results");
  my $wclan;
  my $lclan;
  my $tie;
  my $resstr;

  $tie = ($winner eq $loser);

  if ($tie) {
      $winner = "blue";
      $loser = "red";
  }

  if ($clans_detected) {
      $wclan = " (\U$team2clan{$defteams{$winner}})" if ($team2clan{$defteams{$winner}});
      $lclan = " (\U$team2clan{$defteams{$loser}})" if ($team2clan{$defteams{$loser}});
  }

  my $winnerstr = "<font color=${teamcolors[$defteams{$winner} - 1]}>$winner$wclan</font>";
  my $loserstr = "<font color=${teamcolors[$defteams{$loser} - 1]}>$loser$lclan</font>";

  if (! $tie) {
      $resstr = "<b> $winnerstr </b> defeated $loserstr";
  } else {
      $resstr = "$winnerstr and $loserstr tied";
  }
  
  for ($i = 1; $i <= $maxteam; $i++) {
    $restables .= "<td><table border=1>
	<tr><th colspan=2><font color=${teamcolors[$i - 1]}>Team $i";
    $restables .= " (\U$team2clan{$i})" if $clans_detected;
    $restables .= "</font></th></tr>";
    $restables .= "<tr><td>Players:</td><td>$results_players{$i}</td></tr>" 
	if (keys(%results_players));
    $restables .= "<tr><td>Frags (Drop):  </td><td>$results_frags{$i} ($results_fragsdropped{$i})</td></tr>" if (keys(%results_frags));
    $restables .= "<tr><td>Score:  </td><td>$results_score{$i} </td></tr>"
	if (keys(%results_score));
    $restables .= "</table></td>\n";

  }
  
  print eval "sprintf \"$result_template\n\"";
}

sub print_playerlist {
  my $numplayers = @goodplayers;
  
  my $playerlist = join "<br>\n", map {sprintf("<a href=\"#%s\" >%s</a>", clean_name($_, 1), $_);} @goodplayers;
  
  print eval "sprintf \"$playerlist_template\n\"";
}

sub print_player_summary {
  my @splayers = @goodplayers;
  
  my $numplayers = @splayers;
  my $statslist;
  my $i;
  my $statentries;

  my $stitle = generate_title("Player Summary");
  
  my $statheaders = "\"Name\", \"Rank\", \"Efficiency\", \"Kills\", \"Deaths\", \"Suicides\", \"Team Kills\""; 
  my $statformat = "\"string\", \"int_nototal\", \"percent\", \"int\", \"int\", \"int\", \"int\"";
  foreach (@splayers) {
    $statentries = join(", ", "\"$_\"", worth($_), efficiency($_), 
			$kills{$_} || 0, $deaths{$_} || 0, $suicides{$_} || 0, 
			$killedbuddy{$_} || 0);
    $statslist .= sprintf("indiv_stats[%d] = new Array(%s);\n", $i++, $statentries);
  }
  
  print eval "sprintf \"$playersum_template\n\"";
}




sub print_playerkills {
  my @splayers = @goodplayers;
  
  my $playercols;
  my ($i, $j);
  for ($j = 0; $j < @splayers; $j++) {
    $playercols .= sprintf("<th>%d</th>", $j + 1);
  }
  
  my $playerrows;
  
  my $pi;

  my $stitle = generate_title("Kills");
  
  for ($i = 0; $i < @splayers; $i++) {
    ($pi = $splayers[$i]);
    $playerrows .=  "<tr>";
    $playerrows .=  sprintf("<th align=left><font size=-1>%d&nbsp;$pi</font></th>\n", $i + 1);
    for ($j = 0; $j < @splayers; $j++) {
      $playerrows .=  "<td align=right>";
      if ($i == $j) {
	$playerrows .=  sprintf("<font color=red>%s</font>", 
				$killedby{$splayers[$j], 
					  $splayers[$i]} || "0");
      } else {
	$playerrows .=  $killedby{$splayers[$j], $splayers[$i]} || "0";
      }
      $playerrows .=  "</td>";
    }
    $playerrows .=  "<td align=right>$kills{$splayers[$i]}</td>" .
      "<td align=right>$deaths{$splayers[$i]}</td>" .
	"<td align=right>" .
	  (dnp($splayers[$i]) ? "" : worth($splayers[$i])) 
	    . "</td>\n";
    $playerrows .=  "</tr>\n";
  }
  
  my $deathrow;
  for ($j = 0; $j < @splayers; $j++) {
    $deathrow .= "<td>$bfonth$deaths{$splayers[$j]}$bfontt</td>";
  }
  my $totaldeaths = 0;
  my $totalkills = 0;
  foreach (@splayers) {
    $totaldeaths += $deaths{$_};
    $totalkills += $kills{$_};
  }
  
  print eval "sprintf \"$playerkills_template\n\"";
  
}

sub print_player_indiv {
  my ($player) = @_;
  my @weapons = sort keys %weaponkills;
  my $worth = worth($player);
  my $efficiency = effstr($player);
  # force these to int
  my $kills = $kills{$player} + 0;
  my $deaths = $deaths{$player} + 0;
  my $klutz = $klutz{$player} + 0;
  my $suicides = $suicides{$player} + 0;
  my $killedbuddy = $killedbuddy{$player} + 0;
  my $touches = $flagtouch{$player} + 0;
  my $captures = $flagcaps{$player} + 0;
  my $fumbles = $flagdrops{$player} + 0;
  my $biodiseased = $biodiseased{$player} + 0;  
  
  my $killedwithweaps = "";
  my $killedbyweaps = "";
  my $playerref = clean_name($player, 1);
  
  foreach (@weapons) {
    if ($killedbyweap{$player, $_}) {
      $killedbyweaps .= "<br>" . capweap($_) . 
	": " . $killedbyweap{$player, $_} .
	  sprintf(" (%.1f%%%%)\n", 
		  100.0 * ($killedbyweap{$player, $_} / $deaths{$player}));
    }
  }
  
  foreach (@weapons) {
    if ($killedwithweap{$player, $_}) {
      $killedwithweaps .= "<br>" . capweap($_) . 
	": ". $killedwithweap{$player, $_} .
	  sprintf(" (%.1f%%%%)\n", 
		  100.0 * ($killedwithweap{$player, $_} / $kills{$player}));
    }
  }
  print eval "sprintf \"$indiv_template<p>\n\"";
}

sub print_player_table {
  my($title, $rfunc, $sfunc, $rname, 
     $highaward, $highawardstr, $lowaward, $lowawardstr, 
     $total, $useall, $hidden) = @_;
  
  my $name = $rname;
  $name =~ s/ /_/g;
  push(@sections, $name) unless $hidden;
  
  my @validplayers;
  foreach (@goodplayers) {
    push(@validplayers, $_) if ($useall || &$rfunc($_));
  }
  my @splayers = reverse sort { &$sfunc() } @validplayers;
  return unless (@splayers);
  
  my $rows;
  my $totalrow;
  
  my $r;
  my $sum = 0;
  foreach (@splayers) {
    $r = &$rfunc($_);
    $sum += $r if ($total);
    $rows .= "<tr><td>$_</td><td align=right>$r</td></tr>\n"
      if ($r && !dnp($_));
  }
  $totalrow = "<tr><th>${bfonth}Total$bfontt</th><td align=right>$ofonth$sum$ofontt</td></tr>\n" if ($total);

  print eval "sprintf \"$table_template<p>\n\"" unless($hidden);
  
  if (@splayers) {
    my @highmatch = get_matches($rfunc, @splayers);
    my @lowmatch = get_matches($rfunc, reverse @splayers);
    
    $award{$highaward} = sprintf($highawardstr, join(", ", @highmatch),
				 &$rfunc($highmatch[0])) 
      if ($highaward ne "");
    $award{$lowaward} = sprintf($lowawardstr, join(", ", @lowmatch),
				&$rfunc($lowmatch[0])) 
      if ($lowaward ne "");
  }
  
}


sub print_worthy_players {
    my ($hide) = @_;
  print_player_table("Most Worthy<br><font size=-2>(Kills - Deaths)</font>", 
		     \&worth, \&byworth, "Rank", 
		     "a Best Fragger", 
		     "%s with a killing rank of %d.", 
		     "g Quake Newbie Award", 
		     "%s with a killing rank of %d.", 0, 1, $hide);
  
}

sub print_efficient_players {
    my ($hide) = @_;
  print_player_table("Most Efficient<br><font size=-2>Kills / (Kills + Deaths)</font>", 
		     \&effstr, \&byeff, "Efficiency", 
		     "f Terminator Award", 
		     "%s with a killing efficiency of %d.", 
		     "", "", 0, 1, $hide); 
}

sub print_player_kills {
    my ($hide) = @_;
  print_player_table("Top Killers", sub {my ($f) = @_; $kills{$f}}, sub {$kills{$a} <=> $kills{$b}}, "Kills", 
		     "d Slaughterhouse Award", 
		     "%s with a total of %d kills.", 
		     "h Mr. Magoo Award", 
		     "%s for killing only %d moving objects.", 1, 1, $hide);
}

sub print_player_deaths {
    my ($hide) = @_;
  print_player_table("Top Killees", sub {my ($f) = @_; $deaths{$f}}, sub {$deaths{$a} <=> $deaths{$b}}, "Deaths", 
		     "i Slaughterhouse Llama Award", 
		     "%s getting slaughtered a total of %d times.", 
		     "e Sly Fox Award", 
		     "%s for getting killed only %d times.", 1, 1, $hide);
  
}

sub print_touched {
  print_player_table("Touched", sub {my ($f) = @_; $flagtouch{$f}}, sub {$flagtouch{$a} <=> $flagtouch{$b}}, "Touches", 
		     "b Blowing by the D Award", 
		     "%s for touching the flag %d times.", 
		     "", "", 1);
}

sub print_returned {
  print_player_table("Returned", sub {my ($f) = @_; $flagreturns{$f}}, sub {$flagreturns{$a} <=> $flagreturns{$b}}, "Saves", 
		     "ea Captain Denial", 
		     "%s for returning the flag %d times.", 
		     "", "", 1);
}

sub print_assists {
  print_player_table("Assisted", sub {my ($f) = @_; $assist{$f}}, sub {$assist{$a} <=> $assist{$b}}, "Assists", 
		     "fa John Stockton Award", 
		     "%s for assisting with captures %d times.", 
		     "", "", 1);
}

sub print_defends {
  print_player_table("Defended", sub {my ($f) = @_; $defend{$f}}, sub {$defend{$a} <=> $defend{$b}}, "Defended", 
		     "fb Spank Squad Award", 
		     "%s for outstanding defense %d times", 
		     "", "", 1);
}

sub print_bonus {
  print_player_table("Intercepted", sub {my ($f) = @_; $bonus{$f}}, sub {$bonus{$a} <=> $bonus{$b}}, "Ints", 
		     "fc The Chad Brown Award", 
		     "%s for killing the flag carrier %d times.", 
		     "", "", 1);
}



sub print_capped {
  print_player_table("Capped", sub {my ($f) = @_; $flagcaps{$f}}, sub {$flagcaps{$a} <=> $flagcaps{$b}}, "Caps", 
		     "c The Jerome Bettis Award", 
		     "%s for capping the flag %d times.", 
		     "", "", 1);
}

sub print_fumbled {
  print_player_table("Fumbled", sub {my ($f) = @_; $flagdrops{$f}}, sub {$flagdrops{$a} <=> $flagdrops{$b}}, "Fumbles", 
		     "l The Dave Krieg \"Where'd dat ball go?\" Award", 
		     "%s for fumbling the flag %d times.", 
		     "", "", 1);
}


sub print_dropped {
  print_player_table("Dropped", sub {my ($f) = @_; $dropped{$f}}, sub {$dropped{$a} <=> $dropped{$b}}, "Drops", 
		     "o What's this button do? Award", 
		     "%s for dropping %d times.", 
		     "", "", 1);
}

sub print_overflowed {
  print_player_table("Overflowed", sub {my ($f) = @_; $overflowed{$f}}, sub {$overflowed{$a} <=> $overflowed{$b}}, "Overflows", 
		     "n Sucky ISP Award", 
		     "%s for overflowing %d times.", "", "", 1);
}

sub print_suicide {
    my ($hide) = @_;
  print_player_table("Suicides", sub {my ($f) = @_; $killedby{$f, $f}}, sub {$killedby{$a, $a} <=> $killedby{$b, $b}}, "Suicides", 
		     "j The Dr. Jack Award", 
		     "%s for committing suicide %d times.", "", "", 1, 0, $hide);
}


sub print_killedbuddy {
    my ($hide) = @_;
  print_player_table("Team Kills", sub {my ($f) = @_; $killedbuddy{$f}}, sub {$killedbuddy{$a} <=> $killedbuddy{$b}}, "Buddies", 
		     "k Friends? We don't need no stinkin' Friends! Award", 
		     "%s for killing %d teammates.", "", "", 1, 0, $hide);
}

sub print_diseased {
    my ($hide) = @_;
  print_player_table("Diseased", sub {my ($f) = @_; $biodiseased{$f}}, sub {$biodiseased{$a} <=> $biodiseased{$b}}, "diseased", 
		     "mc The Sick Monkey Award", 
		     "%s for being an infectious carrier %d times.", "", "", 1, 0, $hide);
}


sub capweap {
  my ($weap) = @_;
  $weap =~ s/_/ /g;
  $weap =~ tr/A-Z/a-z/;
  $weap =~ s/\b([a-z])/\u$1/g;
  return $weap;
}

sub old_print_player_weaponkills {
  my @weapons = sort keys %weaponkills;
  my @splayers = @goodplayers;
  
  foreach (@weapons) {
    my $weap = $_;
    print_player_table(capweap("$weap"), 
		       sub {my ($f) = @_; $killedwithweap{$f, $weap}}, 
		       sub {$killedwithweap{$a, $weap} <=> $killedwithweap{$b, $weap}},
		       "Kills", "", "", "", "", 1);
    print "<p>\n";
    
  }
  
}

sub enquote {
  my @l = @_;
  my @o;
  
  foreach (@_) {
    push @o, "\"$_\"";
  }
  return @o;
}

sub print_player_weaponkills {
  my @splayers = @goodplayers;
  my @weapons = sort keys %weaponkills;
  
  my $numplayers = @splayers;
  my $statslist;
  my $i = 0;
  my $statentries;
  my $s;
  
  my $statheaders = join(", ", ("\"Name\"", map {capweap($_)} enquote(@weapons)));
  my $statformat = "\"string\", " . join(", ", map { "\"int\"" } @weapons);
  foreach $s (@splayers) {
    my @weaponkills;
    foreach (@weapons) {
      push @weaponkills, $killedwithweap{$s, $_} || 0;
    }
    $statentries = join(", ", "\"$s\"", @weaponkills);
    $statslist .= sprintf("weaponkills_stats[%d] = new Array(%s);\n", $i++, $statentries);
  }
  print eval "sprintf \"$weaponkills_template\n\"";
}

sub print_player_classkills {
  my ($class) = @_;
  my @splayers;
  my @weapons = sort split(/\s/, $weapons{$class});
  
  my %totalkills;
  my ($p, $w);
  foreach $p (@goodplayers) {
    foreach $w (@weapons) {
      $totalkills{$p, $class} += $killedwithweapclass{$p, $w, $class};
    }
    push @splayers, $p if ($totalkills{$p, $class});
  }
  
  my $numplayers = @splayers;
  my $statslist;
  my $i = 0;
  my $statentries;
  my $s;
  
  my $statheaders = join(", ", ("\"Name\"", map {capweap($_)} enquote(@weapons)));
  my $statformat = "\"string\", " . join(", ", map { "\"int\"" } @weapons);
  foreach $s (@splayers) {
    my @classkills;
    foreach (@weapons) {
      push @classkills, $killedwithweapclass{$s, $_, $class} || 0;
    }
    $statentries = join(", ", "\"$s\"", @classkills);
    $statslist .= sprintf("${class}kills_stats[%d] = new Array(%s);\n", $i++, $statentries);
  }
  print eval "sprintf \"$classkills_template\n\"";
}


sub print_player_weapondeaths {
    my ($hide) = @_;
  my @weapons = sort keys %weaponkills;
  my @splayers = @goodplayers;
  
  foreach (@weapons) {
    my $weap = $_;
    print_player_table(capweap("$weap"), 
		       sub {my ($f) = @_; $killedbyweap{$f, $weap}}, 
		       sub {$killedbyweap{$a, $weap} <=> $killedbyweap{$b, $weap}},
		       "Deaths", "", "", "", "", 1, 0, $hide);
    print "<p>\n";
    
  }
  
}

sub print_randoms {
  
  print_player_table("Talkers",
		     sub {my ($f) = @_; $chatty{$f}}, 
		     sub {$chatty{$a} <=> $chatty{$b}},
		     "Msgs", 
		     "m The Blabbermouth", 
		     "%s for talking %d times during the match.", 
		     "", "", 1) if (keys %chatty);
  
  print "<p>\n";
  
  print_player_table("Death from Natural Causes<br><font size=-1>(Fell to death, drowned, etc)</font>",
		     sub {my ($f) = @_; $klutz{$f}}, 
		     sub {$klutz{$a} <=> $klutz{$b}},
		     "Deaths", 
		     "", 
		     "", 
		     "", "", 1) if (keys %klutz);
  
  print "<p>\n";
  
  print_player_table("Death from Monsters",
		     sub {my ($f) = @_; $monster{$f}}, 
		     sub {$monster{$a} <=> $monster{$b}},
		     "Deaths", 
		     "", 
		     "", 
		     "", "", 1) if (keys %monster);
  
  print "<p>\n";
  
  print_player_table("Death from Traps<br><font size=-1>(Squished, spiked, etc.)</font>",
		     sub {my ($f) = @_; $trap{$f}}, 
		     sub {$trap{$a} <=> $trap{$b}},
		     "Deaths", 
		     "", 
		     "", 
		     "", "", 1) if (keys %trap);
  
  print "<p>\n";
  
  print_player_table("Built Sentry / Dispenser",
		     sub {my ($f) = @_; $build{$f}}, 
		     sub {$build{$a} <=> $build{$b}},
		     "Things", 
		     "n The Dilbert Award", 
		     "%s for building %d sentries / dispensers<br><font size=-1>(What, couldn't get it right the first time?)</font>", 
		     "", "", 1) if (keys %build);
  
  print "<p>\n";
  
  print_player_table("Threw the Pin, not the Grenade",
		     sub {my ($f) = @_; $notoss{$f}}, 
		     sub {$notoss{$a} <=> $notoss{$b}},
		     "Oops", 
		     "", 
		     "", 
		     "", "", 1) if (keys %notoss);

  print "<p>\n";

  print_player_table("Spy Unmaskers",
		     sub {my ($f) = @_; $caught_spy{$f}}, 
		     sub {$caught_spy{$a} <=> $caught_spy{$b}},
		     "Spies", 
		     "fa Goldfinger", 
		     "%s for unmasking %d enemy spies.", 
		     "", "", 1) if (keys %caught_spy);
  
  print "<p>\n";

  print_player_table("Transmitted Infection",
		     sub {my ($f) = @_; $bioinfect{$f}}, 
		     sub {$bioinfect{$a} <=> $bioinfect{$b}},
		     "Infected", 
		     "ma The Ebola Award", 
		     "%s for spreading disease to %d enemies.", 
		     "", "", 1) if (keys %bioinfect);

  print "<p>\n";

  print_player_table("Cured<br><font size=-1>(Infection, Concussion, Gas, Fire)</font>",
		     sub {my ($f) = @_; $biocured{$f}}, 
		     sub {$biocured{$a} <=> $biocured{$b}},
		     "Cured", 
		     "mb The Florence Nightingale Award", 
		     "%s for curing %d teammates.", 
		     "", "", 1) if (keys %biocured);

  print "<p>\n";
  
}


sub print_player_rankings {
    my ($hide) = @_;
    my @splayers = @goodplayers;

    if (! $hide) {  
#  print "$layerh\n";

	print "<NOLAYER>\n";
	print "<center>";
	print generate_title("Player Rankings");
	print "<table border=0 width=\"100%\" cellspacing=10>\n";
	
	print "<tr>";
	print "<td align=center valign=top>";
    }
    print_worthy_players($hide);
    if (! $hide) {
	print "</td>\n";
	print "<td align=center valign=top>";
    }
    print_efficient_players($hide);
    if (! $hide) { 
	print "</td>\n";
	print "</tr>\n";
	
	print "<tr>";
	print "<td align=center valign=top>";
    }
    print_player_kills($hide);
    if (! $hide) {
	print "</td>\n";
	print "<td align=center valign=top>";
    }
    print_player_deaths($hide);
    if (! $hide) {
	print "</td>\n";
	print "</tr>\n";
	
	
	print "<tr>";
	print "<td align=center valign=top>";
    }

    print_diseased($hide);
    if (! $hide) {
	print "</td>\n";
	print "<td align=center valign=top>";
    }
#    print_efficient_players($hide);
    if (! $hide) { 
	print "</td>\n";
	print "</tr>\n";
	
	print "<tr>";
	print "<td align=center valign=top>";
    }

    print_suicide($hide);
    if (! $hide) {
	print "</td>\n";
	print "<td align=center valign=top>";
    }
    print_killedbuddy($hide) if (%killedbuddy != ()) ;
    if (! $hide) {
	print "</td>\n";
	print "</tr>\n";
	
	print "<tr>\n";
	print "<td colspan=2 align=center valign=top>";
	print generate_title("The Weapon Targets");
	print "<multicol cols=3>";
    }
    print_player_weapondeaths($hide);
    if (! $hide) {
	print "</multicol>";
	print "</td>";
	print "</tr>\n";
	
	print "</table>\n";
	print "</center>\n";
	print "</NOLAYER>\n";
#  print "$layert\n";
    }  
}

sub print_flagrunning {
    my $print_something = 0;  

# hmm.... ya wonder if just scalar works...
    $print_something++ if (scalar(keys(%flagtouch  )));
    $print_something++ if (scalar(keys(%flagcaps   )));
    $print_something++ if (scalar(keys(%flagdrops  )));
    $print_something++ if (scalar(keys(%flagreturns)));

    return unless $print_something;

  print "$layerh\n";
  print "<p>\n";
  print "<a name=\"FlagRunning\">\n";
  print "<center>";
  print generate_title("Flag Running");
    print "<font size=-1>\n" if ($print_something > 3);
  print "<table border=0 width=\"100%\" cellspacing=10>\n";
  
    print "<tr>\n";


    if (keys(%flagtouch)) {
	print "<td valign=top align=center>\n";
	print_touched();
	print "</td>\n";

    }

    if (keys(%flagdrops)) {
	print "<td valign=top align=center>\n";
	print_fumbled();
	print "</td>\n";

    }
    if (keys(%flagreturns)) {
	print "<td valign=top align=center>\n";
	print_returned();
	print "</td>\n";

    }
    if (keys(%flagcaps)) {
	print "<td valign=top align=center>\n";
	print_capped();
	print "</td>\n";

    }

    if (%assist || %defend || %bonus) {
	print "</tr><tr>\n";

	if (%assist) {
	    print "<td valign=top align=center>\n";
	    print_assists();
	    print "</td>\n";
	}

	if (%defend) {
	print "<td valign=top align=center>\n";
	print_defends();
	print "</td>\n";

    }
    if (%bonus) {
	print "<td valign=top align=center>\n";
    print_bonus();
	print "</td>\n";

    }


    }

    if (! $print_something) {
	print "<td valign=top align=center>\n";
	print "No Flag Running information available for this log.\n";
	print "</td>\n";
    }

  print "</tr>\n";
  
  print "</table>\n";
    print "</font>\n" if ($print_something > 3);
  print "</center>\n";
  print "$layert\n";
  
}


sub print_player_randoms {
  my @splayers = @goodplayers;
  
  print "$layerh\n";
  print "<p>\n";
  print "<a name=\"Random\">\n";
  print "<center>";
  print generate_title("Random Bits");
  print "<table border=0 width=\"100%\" cellspacing=10>\n";
  
  print "<tr>\n";
  print "<td colspan=2 align=center valign=top>";
  print "<multicol cols=3>";
  print_randoms();
  print_dropped();
  print_overflowed();
  print "</multicol>";
  print "</td>";
  print "</tr>\n";
  
  print "</table>\n";
  print "</center>\n";
  print "$layert\n";
  
}

sub worth {
  my ($targ) = @_;
  return -100 if dnp($targ);
  return $kills{$targ} - $deaths{$targ};
}

sub efficiency {
  my ($targ) = @_;
  return -1 if dnp($targ);
  return sprintf("%.3f", $kills{$targ} / ($kills{$targ} + $deaths{$targ}));
}

sub effstr {
  my ($targ) = @_;
  my $eff = efficiency($targ);
  return sprintf("%d%%%%", $eff * 100.0);
}

sub dnp {
  my ($targ) = @_;
  # if you didn't kill, and you didn't die, you didn't do jack!
  return ($kills{$targ} + $deaths{$targ} == 0);
}

sub byworth {
  my $r = 0 + (worth($a) <=> worth($b));
  #    print STDERR "Comparing $a and $b ($r)\n";
  return $r;
}

sub byeff {
  return efficiency($a) <=> efficiency($b);
}

sub bydeath {
  return $deaths{$a} <=> $deaths{$b};
}

sub bykills {
  return $kills{$a} <=> $kills{$b};
}

sub print_footer {
  print eval "sprintf \"$foot_template<p>\n\"";
}

sub parse_logfile {
    my ($logfile) = @_;
    
    my $targ;
    my $oline;
    my $dtype;
    my $i;
    my $reg;
    my $att;
    my $line;

    my ($pre, $match, $post, $last);

    my $logfd;

    if ($logfile eq "-") {
	$logfd = \*STDIN;
    } else {
	print "Trying to open $logfile\n";
	$logfd = new IO::File;
	open($logfd, "<$logfile") || die("Couldn't open logfile $logfile!"); 
#	$logfd = 'BLARP';
    }
    open(OUTLOG, ">$logfile-tmp") || die("Couldn't open $logfile-tmp!");
    $parsed = 0;
    my $good = 0;
    my $origline;
    $logline = 0;

#  print "$massivere\n";
    
  PARSE:
    while (<$logfd>) {
# just copy it over to OUTLOG for now...
	print OUTLOG unless ($noprivate && (/^\([^:]*\): *.*/ || /^rcon/i));
	$logline++;
	$dtype = "";
	$targ = "";
	$att = "";
	
	s/\s*$//;	# not as good as chomp, but weird DOSisms exist
	
	next if (/^$/);
	$oline = $_;
	$parsed++;
	$origline = $_;
	$line = $_;

      REPARSE:	
	if ($line =~ /$massivere/i && $line !~ /$talkre/i) {
	    ($pre, $match, $post, $last) = ($`, $&, $', $+);
	    print "Matched (last): $match ($last)\n" if ($testing =~ /[am]/);
	    $targ = $pre;
	    $att = $last;
	    $mmatch = $match;
	    my $escaped = $last;
	    $escaped =~ s/(\W)/\\$1/g;
	    #	    $mmatch =~ s/$1// if ($1);
	    #	    $mmatch =~ s/$2// if ($2);
	    $mmatch =~ s/$escaped// if ($last ne "");
            #       print "hashed: $mmatch ($escaped vs $last)\n";
	    # 	    print "$line matched massivere!\n";
	    # 	    print "pre:  $pre\n";
	    # 	    print "mat:  $match\n";
	    # 	    print "post: $post\n";
	    # 	    print "last: $last\n";
	    # 	    print "hash($mmatch): $massivehash{$mmatch}\n";
	    $mmatch =~ s/ //g;
	    $dtype = $massivehash{"\U$mmatch"};
	    print "MParsed: Type: $dtype\tTarg: $targ\tAtt: $att\t($match) [$logline]\n" if ($testing =~ /[ap]/);
            #      print "massivere is: $massivere\n";
	    if ($testing =~ /[am]/) {
		foreach (@massiverelist) {
		    print "\tMatched $_\n" if ($line =~ /$_/i);
		}
	    }
	} elsif ($line =~ /$ignorere/) {
	    ## IMPORTANT NOTE: ignorere is NOT case insensitive!

	    $dtype = "IGNORE";
	} else {

	    for ($i = 0; $i < $numregexps; $i++) {
		$reg = $regexps[$i];
		#	 print "Checking $reg with $line\n";
		if ($line =~ /$reg/i) {
		    $targ = $1;
		    $att = $2;
		    $dtype = $typeidx[$i];
		    print "IParsed: Type: $dtype\tTarg: $targ\tAtt: $att\n" if ($testing  =~ /[ap]/);
		    last;
		    
		}
	    }
	}
	
	$good++;
	$targ =~ s/ +$//;
	$att =~ s/ +$//;

	my ($otarg, $oatt) = ($targ, $att);

	$targ = $munged_names{$otarg};
	if ($targ eq "") {
	    $targ = munge_name($otarg);

	    $munged_names{$otarg} = $targ;
	    $original_names{$targ} = $otarg;
	    $clean_names{$targ} = munge_name($otarg, 1);
	}
	$att = $munged_names{$oatt};
	if ($att eq "") {
	    $att = munge_name($oatt);

	    $munged_names{$oatt} = $att;
	    $original_names{$att} = $oatt;
	    $clean_names{$att} = munge_name($oatt, 1);
	}

	$_ = $dtype;
	study;
	
      SWITCH: {
	  /^MODE_QUAKE2_LOAD/o && do {
	      my $ppost = $post;
	      my @loads = split(/
                                     
/, $ppost);
	      
	      $ppost =~ s/.*
//g;
	      my $lastload = pop(@loads);
	      print "Q2 found ", $ppost, " (vs $lastload)\n" if ($testing);

# pretty nasty hack we use to extract the game when started from a demo
	      if ($mode !~ /q2/) {
		  # skip models/{items, objects, weapons}
		  # hopefully, we'll just have models/ctf or some such. 
		  # nasty hack, I know...
		  my @poss = grep {/^models\// && ! /^models\/(items|objects|weapons)/} @loads;
		  my %possh = {};
		  map {$possh{$1}++ if /models\/([^\/]*)\//} @poss;
		  my @keys = keys(%possh);
		  if (length(@keys) > 1) {
		      print "ERROR! Couldn't figure out what mode you're in!\n";
		      print "Please run speedstats as:\n";
		      print "$cmdline -mode [ ", join(" | ", @keys), " ]\n";
		      exit(1);
		  } else {
		      if ($mode ne "q2" . $keys[0]) {
			  $mode = "q2" . $keys[0];
			  print "Setting mode to $mode\n";
			  if (-e "$templatedir/${mode}_$classfile") {
			      read_classes("${mode}_$classfile");
			  } else {
			      print("Error, couldn't switch to mode $mode (no ${mode}_$classfile in $templatedir!\n");
			  }
		      }
		  }
	      }
	  

	      $origline = $ppost;
	      $line = $origline;
	      $dtype = "";
	      goto REPARSE;

	  };
	  /^MODE_QUAKE2/o && do {
	      $targ = "q2" . $mode if ($targ !~ /q2/);
	  };
	  /^MODE/o && do {
	      $modeline = $line;
	      if ($targ ne $mode) {
		  if (-e "$templatedir/${mode}_$classfile") {
		      $mode = $targ;
		      read_classes("${mode}_$classfile");
		  } else {
		      print("Error, couldn't switch to mode $mode (no ${mode}_$classfile in $templatedir!\n");
		  }
	      }
	      last;
	  };
	  /^SEPARATOR/o && do {	# we should output whatever we have
	  };
	  /^NEWMAP/o && do {	# output whatever we have and start over
	      print STDERR "\n" unless ($quiet);
	      my %oldteam = %team;
	      my $rv = create_file();
	      clear_all();
	      # keep old team information if create_file didn't do
	      # anything. This is because map resets lose join info
	      %team = %oldteam unless ($rv);
	      $thissess++;
	      $spacedtitle = clean_name($targ . "-$thissess", 0, 1);
	      $title = clean_name($targ . "-$thissess", 0);
	      print "New map: $title\n" if ($testing);
	      # yes, the last line in the old one will be the first in
	      # the new one, but that's OK. We need to make sure we
	      # generate the proper map name.
	      print OUTLOG $line;
	      last;
	  };
	  /^TOUCH/o && do {
	      add_player($targ, $oline, $dtype);
	      $flagtouch{$targ}++;
	      
	      # sometimes, this line doesn't end with a \n. So
	      # we trim the regexp on it and see if there's a
	      # reasonable number of chars after. If yes,
	      # reparse the line. 
	      
	      $line = $post;
	      
	      if ($line && length($line) > 20 ) {
		  print "Trimming line from `$origline' to `$line'\n";
		  $origline = $line;
		  $dtype = "";
		  goto REPARSE;
	      }
	      last;
	      
	  };
	  /^CAP/o && do {
	      add_player($targ, $oline, $dtype);
#      print STDERR "$targ CAPPED!\n";
	      $flagcaps{$targ}++;
	      last;
	  };
	  /^FUMBLED/o && do {
	      add_player($targ, $oline, $dtype);
	      $flagdrops{$targ}++;
	      last;
	  };
	  /^MISC/o && do {		#e.g. targ and att are people
	      last;
	  };
	  /^SUICIDE/o && do {
	      add_player($targ, $oline, $dtype);
	      $deaths{$targ}++;
	      $suicides{$targ}++;
	      $killedby{$targ, $targ}++;
	      last;
	  };
	  /^AT_(.*)/o && do {
	      add_player($targ, $oline, $dtype);
	      add_player($att, $oline, $dtype, $line);
	      mark_death($1, $targ, $att);
	      $bonus{$targ}++ if ($bonus_next_kill);
	      last;
	  };
	  /MM1/o && do {
	      #		      add_player($targ, $oline, $dtype);
	      push(@mm1log, "$logline\t$targ\t$att");
	      $chatty{$targ}++;
	      last;
	  };
	  /MMSPEC/o && $spectalk && do {
	      #		      add_player($targ, $oline, $dtype);
	      push(@mm1log, "$logline\t[SPEC] $targ\t$att");
#	      $chatty{$targ}++;
	      last;
	  };
	  /FRIENDLY/o && do {
	      add_player($targ, $oline, $dtype);
	      $killedbuddy{$targ}++;
	      last;
	  };

	  /^INFECT/o && do {
	      add_player($targ, $oline, $dtype);
	      $bioinfect{$targ}++;
	      add_player($att, $oline, $dtype, $line);
		  $biodiseased{$att}++;
	      last;
	  };
	  /^HEAL_/o && do {
	      add_player($targ, $oline, $dtype);
	      $biocured{$targ}++;
	      last;
	  };

	  /ENVIRODEATH/o && do {
	      add_player($targ, $oline, $dtype);
	      $klutz{$targ}++;
	      $deaths{$targ}++;
	      last;
	  };
	  /MONSTERDEATH/o && do {
	      add_player($targ, $oline, $dtype);
	      $monster{$targ}++;
	      $deaths{$targ}++;
	      last;
	  };
	  /TRAPDEATH/o && do {
	      add_player($targ, $oline, $dtype);
	      $trap{$targ}++;
	      $deaths{$targ}++;
	      last;
	  };
	  /^ENTER/o && do {
	      add_player($targ, $oline, $dtype);
	      $enter{$targ}++;
	      last;
	  };
	  /^CLASS/o && do {
	      add_player($targ, $oline, $dtype);
	      $lastclass{$targ} = "\U$att";
	      last;
	  };
	  /^DROPPED/o && do {
	      add_player($targ, $oline, $dtype);
	      $dropped{$targ}++;
	      last;
	  };
	  /^OVERFLOWED|TIMEDOUT/o && do {
	      add_player($targ, $oline, $dtype);
	      $overflowed{$targ}++;
	      last;
	  };
	  /BUILD/o && do {
	      add_player($targ, $oline, $dtype);
	      $build{$targ}++;
	      last;
	  };
	  /^NOTOSS/o && do {
	      add_player($targ, $oline, $dtype);
	      $notoss{$targ}++;
	      last;
	  };
	  /^DETPACKS/o && do {
	      $detpacks++;
	      last;
	  };
	  /^OUTCOME_MATCH/o && do {
	      $resultsin = 1;
	      last;
	  };
	  /^OUTCOME_TEAM/o && do {
	      $team = $targ;
	      last;
	  };
	  /^OUTCOME_PLAYERS/o && do {
	      $results_players{$team} = $targ;
	      last;
	  };
	  /^OUTCOME_SCORE/o && do {
	      $results_score{$team} = $targ;
	      last;
	  };
	  /^OUTCOME_FRAGS/o && do {
	      $results_frags{$team} = $targ;
	      $results_fragsdropped{$team} = $att;
	      last;
	  };
	  /^OUTCOME_WINNER/o && do {
	      $winner = $targ;
	      $loser = $att;
	      last;
	  };
	  /^OUTCOME_DRAW/o && do {
	      $winner = "";
	      $loser = "";
	      last;
	  };
	  /SPY_DETECT/o && do {
	      add_player($targ, $oline, $dtype);
	      $caught_spy{$targ}++;
	      last;
	  };
	  /NAME_CHANGE/o && do {
	      add_player($targ, $oline, $dtype);
	      add_player($att, $oline, $dtype);
	      $id_name{$name_id{$targ}} = $att;
	      # this COULD be painful... since we use names as a key...
	      last;
	  };
	  /JOIN/o && do {
	      add_player($targ, $oline, $dtype);
	      $att = 1 if ($att =~ /blue/i);
	      $att = 2 if ($att =~ /red/i);
	      $team{$targ} = $att;
	      if ($maxteam < $att) {
		  $maxteam = $att;
		  print "Team is now: $maxteam\n" if ($testing =~ /[at]/);
	      }
	      last;
	  };
	  /IGNORE|MM2|MATCHSTART|LEFT|KICKED_SKINS|KICKED_TEAM|SEPARATOR|CHANGED_TEAMS|MMSPEC|SENTRYGUN_DETONATE/o && do {
	      last;
	  };

	  # CTF isms
	  /^ASSIST/o && do {
	      add_player($targ, $oline, $dtype);
	      $assist{$targ}++;
	      last;
	  };
	  /DEFEND/o && do {
	      add_player($targ, $oline, $dtype);
	      $defend{$targ}++;
	      last;
	  };
	  /BONUS_NEXT/o && do {
	      $bonus_next_kill = 1;
	      last;
	  };
	  /BONUS_LAST/o && do {
	      add_player($targ, $oline, $dtype);
	      $bonus{$last_killer}++;
	  };
	  /BONUS/o && do {
	      add_player($targ, $oline, $dtype);
	      $bonus{$targ}++;
	      last;
	  };
	  /RETURNED/o && do {
	      add_player($targ, $oline, $dtype);
	      $flagreturns{$targ}++;
	      last;
	  };
	  /OUTCOME_WIN_BLUE/o && do {
	      $resultsin = 1;
	      $results_score{1} = $targ;
	      $winner = $teamcolors[0];
	      $loser = $teamcolors[1];
	      $maxteam = 2;
	      last;
	  };
	  /OUTCOME_WIN_RED/o && do {
	      $resultsin = 1;
	      $results_score{2} = $targ;
	      $winner = $teamcolors[1];
	      $loser = $teamcolors[0];
	      $maxteam = 2;
	      last;
	  };
	  /OUTCOME_LOSE_BLUE/o && do {
	      $resultsin = 1;
	      $results_score{1} = $targ;
	      $winner = $teamcolors[1];
	      $loser = $teamcolors[0];
	      $maxteam = 2;
	      last;
	  };
	  /OUTCOME_LOSE_RED/o && do {
	      $resultsin = 1;
	      $results_score{2} = $targ;
	      $winner = $teamcolors[0];
	      $loser = $teamcolors[1];
	      $maxteam = 2;
	      last;
	  };
	  /OUTCOME_CRASHED/o && do {
	      last;
	  };
	  /^PROXY/o && do {
	      last;
	  };
	  
	  do {
	      print "**UNKNOWN: $_ ($line) [$logline]\n" if ($testing =~ /[apu]/);
	      last;	  
	  };
	  
      }				# SWITCH
	print STDERR "." unless ($quiet);
    }				# while

    close($logfd) unless ($logfile eq "-");

    print STDERR "\n" unless ($quiet);

    # we're going to break up our friendly logfile into 
    # chunks we care about and chunks we don't. But keep a copy,
    # just in case. :)
    if (! $testing && -e $logfile) {
	print "Renaming $logfile to $logfile.bak\n";
	rename("$logfile", "$logfile.bak");
    }

    print "Parsed $parsed lines, $good successfully\n";
    create_file() if ($parsed > 0);
    close(OUTLOG);
    unlink("$logfile-tmp"); # this is for the one-line tmp file left after create_file
}

# REDRUM! gotta finish up what folks did (e.g. scouts! :)

sub create_file {
  
  @goodplayers = ();
  my $totalkills = 0;
  foreach (keys(%players)) {
    if (!dnp($_)) {
      push(@goodplayers, $_);
      unify_unknown_kills("", $_);
      $totalkills += $kills{$_};
    }
  }
  @goodplayers = sort {clean_name("\U$a", 1) cmp clean_name("\U$b", 1)} @goodplayers;

  my $numplayers = @goodplayers;
  if ($numplayers == 0) {
    print "Yikes! no players! Aborting!\n" if ($testing);
    return 0;
  } elsif ($totalkills < $kill_threshhold) {
    print "Bah... only $totalkills frags (need $kill_threshhold)! Ignoring...\n" if ($testing);
    return 0;
  } else {
    print "We have $numplayers players...\n" if ($testing);
  }

  my $clanstr = extrapolate_clans();

  $title =~ s/&nbsp;/_/g;
  my $fdir = "stats-$clanstr-${title}_${date}";
  $fdir =~ s/\//_/g;
  $fdir =~ s/[^\w _-]//g;
  $fdir = "$statsdir/$fdir";

  print "Writing $fdir...\n";

  (mkdir("$fdir", 0755) || die("Couldn't make $fdir!")) unless (-d "$fdir");
  chdir("$fdir");

  open(OUTFILE, ">index.html");
  select OUTFILE;
  print_frame();
  close(OUTFILE);
  
#  open(OUTFILE, ">$fname");
  open(OUTFILE, ">dataframe.html");
  select OUTFILE;
  
  print_header();
#  print_playerlist();
  print_matchresults() if ($resultsin);
  print_flagrunning();
  print_playerkills();
  print_player_summary();
#  print_player_weaponkills();
#    select $MAINOUT;
### send this to nowhere, but calc awards
#  open('DEVNULL', ">/dev/null");
#  select 'DEVNULL';

  print_player_rankings(1);

#  select OUTFILE;

  print_player_randoms();
  
  foreach (sort keys %classnames) {
    print_player_classkills($_);
  }
  
  print_player_indivstats();
 
  print_mm1log();
  
  print_awards();
  print_footer();
  
  close(OUTFILE);

  open(OUTFILE, ">navframe.html");
  select OUTFILE;
  print_nav();
  close(OUTFILE);

  select $MAINOUT;

  chdir("..");

  close(OUTLOG);
  my $relog = "${title}";
  $relog = "${relog}-$logfile" if (-e $logfile);
  $relog =~ s/.*\///;
  $relog = "${relog}-${date}";
  $relog =~ s/\//-/g;
  $relog =~ s/[^\w _.-]//;
  rename("$logfile-tmp", "$fdir/$relog") || warn("Couldn't rename $logfile-tmp to $fdir/$relog ($!)\n");

  open(OUTLOG, ">$logfile-tmp") || die("Couldn't open $logfile-tmp!");
  print OUTLOG "$modeline\n"; # make sure we continue with the mode

  return 1;
}


sub print_player_indivstats {
  print "$layerh\n";
  print "<p><center>\n";
  print generate_title("Individual Player Stats");  
  print "</center><p>\n";
  print "$layert\n";
  
  foreach (@goodplayers) {
    print_player_indiv($_);
    print "<P>\n";
  }
  print "$layerh\n";
  print "<spacer type=vertical size=10>\n";
  print "$layert\n";
}

sub print_mm1log {
    my $mm1logstr;
    my ($line, $sayer, $blah);

    my $stitle = generate_title("MM1 Log");

    foreach (@mm1log) {
	($line, $sayer, $blah) = split(/\t/, $_);
	$mm1logstr .= "<tr><td>$line</td><td>$sayer</td><td>$blah</td></tr>\n";
    }

    print eval "sprintf \"$mm1_template<p>\n\"";    
}

sub print_awards {
  print "$layerh\n";
  print "<a name=\"Awards\">\n";
  print "<p><center>\n";
  print generate_title("Awards");
  print "<table cellspacing=10>\n";
  my $award_name; # (trim first word from award)
  foreach (sort keys(%award)) {
      ($award_name = $_) =~ s/^(\w*) //;
      print "<tr><td align=left valign=top>\n";
      print "<font size=5>$bfonth $award_name $bfontt</font>\n";
      print "</td><td valign=top align=left>\n";
      print "$award{$_}\n";
      print "</td></tr>\n";
  }
  print "</table>\n";
  print "</center><p>\n";
  print "$layert\n";
}


sub mark_death {
  my ($weap, $targ, $att) = @_;
  
  my $class;
  my $mode;

  # weapon may have things like mFORTRESS (indicating we're in FORTRESS mode)
  # or cSCOUT, indicating it's the SCOUT version of the weapon
  
  if ($weap =~ /_c([^_]*)/) {
      $class = $1;
      $weap =~ s/$&//;
  }
  if ($weap =~ /_m([^_]*)/) {
      $mode = $1;
      $weap =~ s/$&//;
  }
  

  $last_killer = $att;
  $kills{$att}++;
  $deaths{$targ}++;
  $killedby{$targ, $att}++;
  $killedbyweap{$targ, $weap}++;
  $killedwithweap{$att, $weap}++;
  $weaponkills{$weap}++;
  
  $class = guess_class($weap, $att, $targ) unless ($class);
  if ($class eq "ERROR") {
    #	print "ERROR! guess_class($weap, $att) had no possibilities!\n";
  } elsif ($class eq "UNKNOWN") {
    # d'oh... couldn't figure it out...
  } else {
    $killedbyweapclass{$targ, $weap, $class}++;
    $killedwithweapclass{$att, $weap, $class}++;
  }
  
}

sub add_player {
  my($targ, $line, $dtype, $mline) = @_;
  
  $players{$targ}++;

  if ($players{$targ} == 1) {  
    my $cn = clean_name($targ, 0, 1);
    print "New Player: $cn at ($dtype) $line [$logline]\n$mline" if ($testing =~ /[ac]/);
    $id_name{scalar(keys(%name_id))} = $targ;
    $name_id{$targ} = scalar(keys(%name_id));

# clan heuristic.... we'll say that <=5 alpha chars == clan, at front or end. 
# we'll assume that 3+ means it is a clan.
    my $clan;
    if ($cn =~ /^\W*(\w{1,5})\W/) {
      ($clan = $1) =~ tr/[A-Z]/[a-z]/;
      $possible_clans{$clan}++;
      $player_clans{$targ, "left"} = $clan;
      push(@{ $clanlist{$clan} }, $targ);
    }
    if ($cn =~ /\W(\w{1,5})\W*$/) {
      ($clan = $1) =~ tr/[A-Z]/[a-z]/;
      $possible_clans{$clan}++;
      $player_clans{$targ, "right"} = $clan;
      push(@{ $clanlist{$clan} }, $targ);
    }
  }
}

sub read_obits {
  my($obitfile) = @_;
  my $dtype;
  my $reg;
  my $mungedreg;
  
  open(OBITFILE, "<$templatedir/$obitfile") 
    || die("Couldn't open obit file ($templatedir/$obitfile)!");
  
  while (<OBITFILE>) {
    chomp;
    next if (/^\#|^$/);		 #
    if (/^([\w]*)\s*\/(.*)\/$/) {
      $dtype = $1;
      $reg = $2;
      if ($dtype eq "IGNORE") {
	$ignorere = (! $ignorere ) ? "$reg" : "$ignorere|$reg";
	print "Adding $reg to ignorere.\n" if ($testing =~ /[ah]/);
      } elsif ($reg =~ /^\^\(\.\*\)/) {
	# if the pattern is /^(.*) blah (.*) blah/
	# we can optimize.
	
	$mungedreg = substr($reg, length("^(.*)"));
	
	$massivere = (! $massivere) ? "$mungedreg" 
	  : "$massivere|$mungedreg";
	
	$mungedreg =~ s/\(\.\*\)//;
	$mungedreg =~ s/\$$//;
	$mungedreg =~ s/\\//g;
	$mungedreg =~ s/ //g;
	print "Adding $mungedreg to massivehash.\n" if ($testing =~ /[ah]/);

	print "Warning: duplicate entry for $mungedreg; setting to $dtype (was ", $massivehash{"\U$mungedreg"}, ")\n"
	    if ($massivehash{"\U$mungedreg"});
	$massivehash{"\U$mungedreg"} = $dtype;


	push(@massivereidx, $dtype);
	push(@massiverelist, $reg);
	
    } elsif ($dtype =~ /^MM/) {
	$talkre = (! $talkre ) ? "$reg" : "$talkre|$reg";
	print "Adding $reg to talkre.\n" if ($testing =~ /[ah]/);

	study $reg;
	push(@typeidx, $dtype);
	push(@regexps, $reg);
    } else {
	study $reg;
	push(@typeidx, $dtype);
	push(@regexps, $reg);
    }
      
      
      #	print "New regexp: $dtype\t$reg\n";
    } else {
	warn "ERROR! Couldn't read line $_\n";
    }
  }
  close(OBITFILE);
  #    print "massivere is:\n\n$massivere\n\n";
  print "Parsed ", $#regexps + 1, " regexps\n";
  $numregexps = @regexps;

  study $massivere;
  study $ignorere;
  study $talkre;
}

sub read_classes {
  my($classfile) = @_;
  my $classstr;
  my $weapon;

  %classes = ();
  %classnames = ();
  %weapons = ();
  
  open(CLASSFILE, "<$templatedir/$classfile") || die("Couldn't open class file ($classfile)!");
  while (<CLASSFILE>) {
    s/\s+$//;
    next if (/^\#|^$/);		 #
    if (/^([\w]*)\s*(.*)$/) {
      $weapon = $1;
      $classstr = $2;
      $classes{$weapon} = "$classstr";
      foreach (split /\s/, $2) {
	$classnames{$_}++ if /\w/;
	$weapons{$_} .= "$weapon ";
      }
      print "New class: \"$weapon\"\t=>\t\"$classes{$weapon}\"\n" if ($testing =~ /[ac]/);
    } else {
      print STDERR "ERROR! Couldn't read line $_\n";
    }
  }
  close(CLASSFILE);
}

######################################################################
## this is a big hairy function that tries to determine what the teams are.
## First, we try to determine clans by \W(\w*)\W at the beginning or
## end of the name (e.g. >V< SPEEDenator or LibTech[OSKI]. We assume
## that clans are alpha-numeric and separated by non-numeric... this
## won't pick up some clans that mix, such as [I'm12]. Oh well.
##
## We then try to use "Joined Team (.*) to determine which team
## they're on. This works for a few people.
##
## Once we know some, we try to extrapolate based on what we
## know... we make the assumption that each team doesn't kill its own
## often (they might 'cuz of telefrags or weird team settings), and
## try to separate them that way.


sub extrapolate_clans {

    foreach (@goodplayers) {
	$setclan{$_} = "";
    }

    print "Extrapolating clan data, team is: $maxteam...\n" if ($testing =~ /[at]/);

    my @possibilities;
    foreach (reverse sort {$possible_clans{$a} <=> $possible_clans{$b}} keys %possible_clans) {
	if ($possible_clans{$_} > 2) {
	    print "\tPossible clan $_: $possible_clans{$_}\n" 
		if ($testing =~ /[at]/);
	    push(@possibilities, $_);
	}	    
    }

    # make clan lists
    my @clist;
    my $clan;
    foreach (@possibilities) {
	@clist = @{ $clanlist{$_} };
	$clan = $_;
	map {$setclan{$_} = $clan} @clist;

	if ($testing =~ /[at]/) {
	    print "Probable members of $clan:\n";
	    print join(", ", map { clean_name($_)} @clist);
	    print "\n";
	}

    }

    # let's make sure we have enough. For most cases, we should be
    # looking at 2 teams. But there's the occasional 3 and 4 team thing there.
    if (scalar(@possibilities) < $maxteam) {
	print "Not enough possible clans for number of teams! (", scalar(@possibilities), " possible, need $maxteam\n" if ($testing =~ /[at]/);
	return;
    }

    # let's try to assign team numbers to clans. Allow some slack for
    # people that accidentally join the wrong team

    my %teammappings;
    foreach (@goodplayers) {
	next unless ($setclan{$_} && $team{$_});  # ignore people we don't know yet
	$teammappings{$setclan{$_}, $team{$_}}++;
    }
    my $t;
    my $unique_mappings = 0;
    my @maplist;

    foreach $t (1 .. $maxteam) {
	foreach (@possibilities) {
	    if ($teammappings{$_, $t}) {
		print "Team Mapping: $_ to Team $t: $teammappings{$_, $t}\n" 
		    if ($testing =~ /[at]/);
		$unique_mappings++;
		push(@maplist, "$teammappings{$_, $t} $_ $t");
	    }
	}
    }
    print "Team Joins: ", scalar(keys(%team)), "\n" if ($testing =~ /[at]/);
    print "Team Mappings: ", scalar(keys(%teammappings)), "\n" if ($testing =~ /[at]/);;
    print "Unique Mappings: $unique_mappings\n" if ($testing =~ /[at]/);;


    # ok... we now have team mappings. If we're lucky, we have 2 and they map
    # perfectly. If we're fortunate, we have one and can deduce the other
    # If life happens, we have lots, and need to pick the first couple...
    # which hopefully will be right. :)
    # if we have only n-2 or fewer mappings (e.g. 2 / 4 or 1/3), we're 
    # screwed (until I can think of some other heuristic to use! :)
    
    # we're missing one.... assume the rest are accurate.

    my $v;
    my $c;

    my $numpos;
    foreach (reverse sort {my @a = split(/ /, $a); my @b = split(/ /, $b); $a[0] <=> $b[0]} @maplist) {
	($v, $c, $t) = split(/ /, $_);
	unless ($team2clan{$t}) {
	    $team2clan{$t} = $c;
	    $clan2team{$c} = $t;
	}
    }

    # we're shy one...
    if (scalar(keys(%clan2team)) == $maxteam - 1) {
	foreach $c (@possibilities) {
	    foreach $t (1 .. $maxteam) {
		unless ($team2clan{$t}) {
		    print "Didn't find mapping for $t, picking $c...\n" if ($testing =~ /[at]/);
		    $team2clan{$t} = $c;
		    $clan2team{$c} = $t;
		}
	    }
	}
    }

    foreach (sort keys(%team2clan)) {
	print "team2clan{$_} = $team2clan{$_}\n" if ($testing =~ /[at]/);
    }

    print "Unknown:\n" if ($testing =~ /[at]/);
    my @unknown;
    foreach (@goodplayers) {
	unless ($setclan{$_}) {
	    print clean_name($_), "\n"  if ($testing =~ /[at]/);
	    push(@unknown, $_);
	}
    }

    my $u;
    foreach $u (@unknown) {
	my %clanmap;
	foreach (@possibilities) {
	    $clanmap{$_} = 0;
	}
	foreach (@goodplayers) {
	    if ($killedby{$_, $u} && $setclan{$_}) {
		$clanmap{$setclan{$_}}++;
	    }
	}
	my @sortmap = sort {$clanmap{$a} <=> $clanmap{$b}} keys %clanmap;
	$setclan{$u} = $sortmap[0];
	print "Setting ", clean_name($u), " clan to: $setclan{$u} ($clanmap{$setclan{$u}})\n" if ($testing =~ /[at]/);
	push(@{ $clanlist{$setclan{$u}} }, $u);
    }

    # sanity check... make sure team1 players aren't killing team2 players, etc.

    my $sanity = 0;
    my $g1;
    my $g2;
    foreach $g1 (@goodplayers) {
	my $curclan = $setclan{$g1};
	foreach $g2 (@goodplayers) {
	    if ($g1 ne $g2 && $killedby{$g2, $g1} 
		&& $setclan{$g2} eq $setclan{$g1}) {
		$sanity += $killedby{$g2, $g1};
		
	    }
	}
    }

    print "Teamcheck Sanity is: $sanity\n" if ($testing =~ /[at]/);
    # teams can kill teams via telefrags...
    print "Telefrags are: $weaponkills{TELEFRAG}\n" if ($testing =~ /[at]/);

    if ($sanity - $weaponkills{"TELEFRAG"} < $teamkill_threshhold) {
	$clans_detected = 1;
#	my @teampos = @possibilities[0 .. ($maxteam - 1)];
	my $rv = join("-", sort keys %clan2team);
	print "Returning $rv ($maxteam)\n" if ($testing =~ /[at]/);
	return $rv;
    }

    return "";
}


sub asciify_name {
    my ($line) = @_;
    
    $line =~ s/\[([0-9a-f][0-9a-f])\]/chr(hex($1))/eg if ($serverlog);

    $line =~ tr/[\000-\037]/ /;
    $line =~ tr/[\200-\237]/ /;
    $line =~ tr/[\240-\377]/[\040-\177]/;
    return $line;
}

sub munge_name {
  my($line, $plain, $keepspaces) = @_;

  my ($ofh, $oft, $gfh, $gft) = ($ofonth, $ofontt, $gfonth, $gfontt);

  if ($plain) {
      ($ofh, $oft, $gfh, $gft) = ("", "", "", "");
  }

  $line =~ s/\[([0-9a-f][0-9a-f])\]/chr(hex($1))/eg if ($serverlog);

  #	$line =~ tr/\0146-\0254/\048-\0126/;
  $line =~ s/ /&nbsp;/g;	# turn spaces into non-blocking spaces... good?
  $line =~ s/\</&lt;/g;
  $line =~ s/\>/&gt;/g;
  $line =~ s/^\002/Mapname: /;
  $line =~ s/\020|\220/$ofh\[$oft/g; #]
  $line =~ s/\021|\221/$ofh\]$oft/g;
  # $line =~ tr [\022-\033] [\060-\071];
  
  my $i;
  my $i2;
  my $rep;
  my $a;
  my $count;
  
  for ($i = 022; $i <= 033; $i++) {
    $a = chr($i);
    next if ($line !~ $a);
    $i2 = $i + 30;
    $rep = $ofh . chr($i2) . $oft;
    $line =~ s/$a/$rep/g;
  }
  # $line =~ tr [\222-\233] [\060-\071];
  for ($i = 0222; $i <= 0233; $i++) {
    $a = chr($i);
    next if ($line !~ $a);
    $i2 = $i - 98;
    $rep = $ofh . chr($i2) . $oft;
    $line =~ s/$a/$rep/g;
  }
  # $line =~ tr [\234-\377] [\034-\177];
  for ($i = 0234; $i <= 0377; $i++) {
    $a = chr($i);
    next if ($line !~ $a);
    $i2 = $i - 0200;
    my $c2 = chr($i2);
    $c2 = "&gt;" if (chr($i2) eq ">");
    $c2 = "&lt;" if (chr($i2) eq "<");
    $rep = $ofh . $c2 . $oft;
    $count = ($line =~ s/$a/$rep/g );
  }
  $line =~ s/\036/=/g;
  $line =~ s/[\000-\037]/&#183;/g;
  $line =~ s/[\177-\237]/&#183;/g;
  $line =~ s/\377/&#183;/g;
  $line =~ s/&#183;/$gfh &#183; $gft/g; 
  $line =~ s/$oft$ofh//g unless ($plain);
  $line =~ s/$gft$gfh//g unless ($plain);
#  $line =~ s/><</>&lt;</g;
#  $line =~ s/>></>&gt;</g;

  $line = " " if ($line eq "");
  return $line;
}

# get rid of fancy stuff munge_name adds... i.e. remove <.*>
sub clean_name {
  my ($name, $escape, $keep_spaces) = @_;
  while ($name =~ /<.*>/) {
    $name =~ s/<[^<>]*>//g;
  }
  $name =~ s/&nbsp;/ /g;
  $name =~ s/&[^;]*;/ /g;
  $name =~ s/\s//g unless ($keep_spaces);

  if ($escape) {
      $name =~ s/\W//g;
#      $name =~ s/\</&lt;/g;
#      $name =~ s/\>/&gt;/g;
#      $name =~ s/\?//g;

  }

  return $name;
}

main();

1;
__END__;

# POD stuff can go here
